from Sampling.CrossValidation cimport CrossValidation
import random


cdef class KFoldCrossValidation(CrossValidation):

    def __init__(self, instanceList: list, K: int, seed: int):
        """
        A constructor of KFoldCrossValidation class which takes a sample as an array of instances, a K (K in K-fold
        cross-validation) and a seed number, then shuffles the original sample using this seed as random number.

        PARAMETERS
        ----------
        instanceList : list
            Original sample
        K : int
            K in K-fold cross-validation
        seed : int
            Random number to create K-fold sample(s)
        """
        self.__instanceList = instanceList
        random.seed(seed)
        random.shuffle(instanceList)
        self.__N = len(instanceList)
        self.K = K

    cpdef list getTrainFold(self, int k):
        """
        getTrainFold returns the k'th train fold in K-fold cross-validation.

        PARAMETERS
        ----------
        k : int
            index for the k'th train fold of the K-fold cross-validation

        RETURNS
        -------
        list
            Produced training sample
        """
        cdef int i
        cdef list trainFold = []
        for i in range((k * self.__N) // self.K):
            trainFold.append(self.__instanceList[i])
        for i in range(((k + 1) * self.__N) // self.K, self.__N):
            trainFold.append(self.__instanceList[i])
        return trainFold

    cpdef list getTestFold(self, int k):
        """
        getTestFold returns the k'th test fold in K-fold cross-validation.

        PARAMETERS
        ----------
        k : int
            index for the k'th test fold of the K-fold cross-validation

        RETURNS
        -------
        list
            Produced testing sample
        """
        cdef int i
        cdef list testFold = []
        for i in range((k * self.__N) // self.K, ((k + 1) * self.__N) // self.K):
            testFold.append(self.__instanceList[i])
        return testFold



cdef class MultipleFile:

    def __init__(self, fileList: list):
        self.fileNameList = fileList
        file = open(self.fileNameList[0])
        self.lines = file.readlines()
        file.close()
        self.lineIndex = 0
        self.index = 0

    cpdef str readLine(self):
        cdef str line
        if self.lineIndex == len(self.lines):
            self.index = self.index + 1
            file = open(self.fileNameList[self.index])
            self.lines = file.readlines()
            file.close()
            self.lineIndex = 0
        line = self.lines[self.lineIndex]
        self.lineIndex = self.lineIndex + 1
        return line



from NGram.SimpleSmoothing cimport SimpleSmoothing


cdef class GoodTuringSmoothing(SimpleSmoothing):

    cpdef list __linearRegressionOnCountsOfCounts(self, list countsOfCounts)
    cpdef setProbabilities(self, object nGram, int level)


import random


cdef class StratifiedKFoldCrossValidation(KFoldCrossValidation):

    def __init__(self, instanceLists: list, K: int, seed: int):
        """
        A constructor of StratifiedKFoldCrossValidation class which takes as set of class samples as an array of array of
        instances, a K (K in K-fold cross-validation) and a seed number, then shuffles each class sample using the
        seed number.

        PARAMETERS
        ----------
        instanceLists : list
            Original class samples. Each element of the this array is a sample only from one class.
        K : int
            K in K-fold cross-validation
        seed : int
            Random number to create K-fold sample(s)
        """
        cdef int i
        self.__instanceLists = instanceLists
        self._N = []
        for i in range(len(instanceLists)):
            random.seed(seed)
            random.shuffle(instanceLists[i])
            self._N.append(len(instanceLists[i]))
        self.K = K

    cpdef list getTrainFold(self, int k):
        """
        getTrainFold returns the k'th train fold in K-fold stratified cross-validation.

        PARAMETERS
        ----------
        k : int
            index for the k'th train fold of the K-fold stratified cross-validation

        RETURNS
        -------
        list
            Produced training sample
        """
        cdef int i, j
        cdef list trainFold = []
        for i in range(len(self._N)):
            for j in range((k * self._N[i]) // self.K):
                trainFold.append(self.__instanceLists[i][j])
            for j in range(((k + 1) * self._N[i]) // self.K, self._N[i]):
                trainFold.append(self.__instanceLists[i][j])
        return trainFold

    cpdef list getTestFold(self, int k):
        """
        getTestFold returns the k'th test fold in K-fold stratified cross-validation.

        PARAMETERS
        ----------
        k : int
            index for the k'th test fold of the K-fold stratified cross-validation

        RETURNS
        -------
        list
            Produced testing sample
        """
        cdef int i, j
        cdef list testFold = []
        for i in range(len(self._N)):
            for j in range((k * self._N[i]) // self.K, ((k + 1) * self._N[i]) // self.K):
                testFold.append(self.__instanceLists[i][j])
        return testFold



from NGram.SimpleSmoothing cimport SimpleSmoothing


cdef class TrainedSmoothing(SimpleSmoothing):

    cpdef learnParameters(self, list corpus, int N)
    cpdef double newLowerBound(self, double current, double currentLowerBound,
                      double currentUpperBound, int numberOfParts)
    cpdef double newUpperBound(self, double current, double currentLowerBound,
                      double currentUpperBound, int numberOfParts)
    cpdef train(self, list corpus, object nGram)


from NGram.NGram cimport NGram


cdef class LaplaceSmoothing(SimpleSmoothing):

    def __init__(self, delta=1.0):
        self.__delta = delta

    cpdef setProbabilities(self, object nGram, int level):
        """
        Wrapper function to set the N-gram probabilities with laplace smoothing.

        PARAMETERS
        ----------
        nGram : NGram
            N-Gram for which the probabilities will be set.
        level : int
            height for NGram. if level = 1, If level = 1, N-Gram is treated as UniGram, if level = 2, N-Gram is treated
            as Bigram, etc.
        """
        nGram.setProbabilityWithPseudoCount(self.__delta, level)



from Sampling.CrossValidation cimport CrossValidation

cdef class KFoldCrossValidation(CrossValidation):

    cdef list __instanceList
    cdef int __N

    cpdef list getTrainFold(self, int k)
    cpdef list getTestFold(self, int k)


cdef class TrainedSmoothing(SimpleSmoothing):

    cpdef learnParameters(self, list corpus, int N):
        pass

    cpdef double newLowerBound(self, double current, double currentLowerBound,
                      double currentUpperBound, int numberOfParts):
        """
        Calculates new lower bound.

        PARAMETERS
        ----------
        current : float
            current value.
        currentLowerBound : float
            current lower bound
        currentUpperBound : float
            current upper bound
        numberOfParts : int
            number of parts between lower and upper bound.

        RETURNS
        -------
        float
            new lower bound
        """
        if current != currentLowerBound:
            return current - (currentUpperBound - currentLowerBound) / numberOfParts
        else:
            return current / numberOfParts

    cpdef double newUpperBound(self, double current, double currentLowerBound,
                      double currentUpperBound, int numberOfParts):
        """
        Calculates new upper bound.

        PARAMETERS
        ----------
        current : float
            current value.
        currentLowerBound : float
            current lower bound
        currentUpperBound : float
            current upper bound
        numberOfParts : int
            number of parts between lower and upper bound.

        RETURNS
        -------
        float
            new upper bound
        """
        if current != currentUpperBound:
            return current + (currentUpperBound - currentLowerBound) / numberOfParts
        else:
            return current * numberOfParts

    cpdef train(self, list corpus, object nGram):
        """
        Wrapper function to learn parameters of the smoothing method and set the N-gram probabilities.

        PARAMETERS
        ----------
        corpus : list
            Train corpus used to optimize parameters of the smoothing method.
        nGram : NGram
            N-Gram for which the probabilities will be set.
        """
        self.learnParameters(corpus, nGram.getN())
        self.setProbabilitiesGeneral(nGram)



cdef class SimpleSmoothing:

    cpdef setProbabilities(self, object nGram, int level):
        pass

    cpdef setProbabilitiesGeneral(self, object nGram):
        self.setProbabilities(nGram, nGram.getN())



from NGram.NGramNode cimport NGramNode
from NGram.SimpleSmoothing cimport SimpleSmoothing
from NGram.TrainedSmoothing cimport TrainedSmoothing


cdef class NGram:

    cdef NGramNode rootNode
    cdef int __N
    cdef double __lambda1, __lambda2
    cdef bint __interpolated
    cdef set __vocabulary
    cdef list __probabilityOfUnseen

    cpdef merge(self, NGram toBeMerged)
    cpdef int getN(self)
    cpdef setN(self, int N)
    cpdef addNGramSentence(self, list symbols, int sentenceCount = *)
    cpdef addNGram(self, list symbols)
    cpdef int vocabularySize(self)
    cpdef setLambda2(self, double lambda1)
    cpdef setLambda3(self, double lambda1, double lambda2)
    cpdef calculateNGramProbabilitiesTrained(self, list corpus, TrainedSmoothing trainedSmoothing)
    cpdef calculateNGramProbabilitiesSimple(self, SimpleSmoothing simpleSmoothing)
    cpdef calculateNGramProbabilitiesSimpleLevel(self, SimpleSmoothing simpleSmoothing, int level)
    cpdef replaceUnknownWords(self, set dictionary)
    cpdef set constructDictionaryWithNonRareWords(self, int level, double probability)
    cpdef double __getUniGramPerplexity(self, list corpus)
    cpdef double __getBiGramPerplexity(self, list corpus)
    cpdef double __getTriGramPerplexity(self, list corpus)
    cpdef double getPerplexity(self, list corpus)
    cpdef double __getUniGramProbability(self, object w1)
    cpdef double __getBiGramProbability(self, object w1, object w2)
    cpdef double __getTriGramProbability(self, object w1, object w2, object w3)
    cpdef int getCount(self, list symbols)
    cpdef setProbabilityWithPseudoCount(self, double pseudoCount, int height)
    cpdef int __maximumOccurence(self, int height)
    cpdef __updateCountsOfCounts(self, list countsOfCounts, int height)
    cpdef list calculateCountsOfCounts(self, int height)
    cpdef setAdjustedProbability(self, list countsOfCounts, int height, double pZero)
    cpdef prune(self, double threshold)
    cpdef saveAsText(self, str fileName)


import random


cdef class Bootstrap:

    def __init__(self, instanceList: list, seed: int):
        """
        A constructor of Bootstrap class which takes a sample an array of instances and a seed number, then creates a
        bootstrap sample using this seed as random number.

        PARAMETERS
        ----------
        instanceList : list
            Original sample
        seed : int
            Random number to create boostrap sample
        """
        cdef int N, i
        random.seed(seed)
        N = len(instanceList)
        self.__instanceList = []
        for i in range(N):
            self.__instanceList.append(instanceList[random.randint(0, N - 1)])

    cpdef list getSample(self):
        """
        getSample returns the produced bootstrap sample.

        RETURNS
        -------
        list
            Produced bootstrap sample
        """
        return self.__instanceList



from DataStructure.CounterHashMap cimport CounterHashMap


cdef class NGramNode(object):

    cdef dict __children
    cdef object __symbol
    cdef int __count
    cdef double __probability, __probabilityOfUnseen
    cdef NGramNode __unknown

    cpdef merge(self, NGramNode toBeMerged)
    cpdef int getCount(self)
    cpdef double getProbability(self)
    cpdef int size(self)
    cpdef int maximumOccurence(self, int height)
    cpdef double childSum(self)
    cpdef updateCountsOfCounts(self, list countsOfCounts, int height)
    cpdef setProbabilityWithPseudoCount(self, double pseudoCount, int height, double vocabularySize)
    cpdef setAdjustedProbability(self, list N, int height, double vocabularySize, double pZero)
    cpdef addNGram(self, list s, int index, int height, int sentenceCount = *)
    cpdef double getUniGramProbability(self, object w1)
    cpdef double getBiGramProbability(self, object w1, object w2)
    cpdef double getTriGramProbability(self, object w1, object w2, object w3)
    cpdef countWords(self, CounterHashMap wordCounter, int height)
    cpdef replaceUnknownWords(self, set dictionary)
    cpdef int getCountForListItem(self, list s, int index)
    cpdef object generateNextString(self, list s, int index)
    cpdef prune(self, double threshold, int N)
    cpdef saveAsText(self, bint isRootNode, object outputFile, int level)


from NGram.SimpleSmoothing cimport SimpleSmoothing


cdef class NoSmoothing(SimpleSmoothing):

    cpdef setProbabilities(self, object nGram, int level):
        nGram.setProbabilityWithPseudoCount(0.0, level)




from Sampling.KFoldCrossValidation cimport KFoldCrossValidation


cdef class StratifiedKFoldCrossValidation(KFoldCrossValidation):

    cdef list __instanceLists
    cdef list _N

    cpdef list getTrainFold(self, int k)
    cpdef list getTestFold(self, int k)
